indexing

   description: "Class that covers the Standard C stream%
   %code. Filenames are subject to expansion."

   idea: "Mimicks an existing file. Tries to maintain that file exists%
   %during life of object by keeping it open from creation to%
   %destruction of the object. But you may close it earlier."

   author: "Berend de Boer"
   date: "$Date: Sat, 14 Oct 2000 08:59:36 +0200 $";
   revision: "$Revision: 1.9 $"
   

deferred class 

   STDC_FILE

inherit
   
   MEMORY
      redefine
         dispose
      end

   STDC_BASE

   PORTABLE_PATH

   CAPI_STDIO

   
feature -- creation

   create_read_write (path: STRING) is
         -- Open file for update (reading and writing). If the file 
         -- already exists, it is truncated to zero length.
         -- So permissions seem to remain.
      do
         do_make
         do_open (path, "w+")
      end

   create_write (path: STRING) is
         -- create new file for writing. If the file already exists, 
         -- it is truncated to zero length.
         -- So permissions seem to remain.
      do
         do_make
         do_open (path, "w")
      end
   
   open (path, a_mode: STRING) is
         -- open file in given mode
      do
         do_make
         do_open (path, a_mode)
      end

   open_append (path: STRING) is
         -- append to exiting file or create file if it does not exist
      do
         do_make
         open (path, "a")
      end
   
   open_read (path: STRING) is
         -- open file for reading
      do
         do_make
         open (path, "r")
      end
   
   open_read_write (path: STRING) is
         -- open file for reading and writing
      do
         do_make
         open (path, "r+")
      end


feature -- work with existing streams
   
   attach_to_stream (a_stream: POINTER; a_mode: STRING) is
         -- attach to `a_stream'. Will become owner of stream so
         -- it will close it when garbage collected.
      require
         valid_stream: a_stream /= Void
         valid_mode: a_mode /= Void and then a_mode.count > 0
         -- `a_stream' is open
         -- `a_mode' is compatible with `a_stream'
      do
         do_make
         stream := a_stream
         set_mode (a_mode)
      end

   unattach is
         -- assume someone else will close this stream
      do
         stream := default_pointer
         mode := Void
      end

   
feature -- close
   
   close is
      local
         r: INTEGER
      do
         r := posix_fclose (stream)
         stream := default_pointer
         if r /= 0 then
            raise_posix_error
         end
      ensure
         closed: not is_open
      end


feature -- reopen
   
   reopen (path, a_mode: STRING) is
         -- closes and then opens a stream
      require
         open: is_open
         -- valid_mode: mode is a valid posix mode
      local
         cpath: POINTER
         cmode: POINTER
      do
         portable_path.make_expand (path)
         cpath := sh.string_to_pointer (portable_path)
         set_mode (a_mode)
         cmode := sh.string_to_pointer (mode)
         stream := posix_freopen (cpath, cmode, stream)
         if not is_open then
            raise_posix_error
         end
      ensure
         file_stays_open: is_open
      end


feature -- control over buffering

   flush is
         -- Updates this stream
      local
         i: INTEGER
      do
         i := posix_fflush (stream)         
         if i /= 0 then
            raise_posix_error
         end
      end

   setbuf, set_buffer (buffer: POINTER) is
         -- Determines how the stream will be buffered
         -- gives you a fully buffered input and output
         -- Not sure: buffer should have at least BUFSIZ bytes?
      require
         -- no operation has yet been performed on this file
         -- `buffer' maybe NULL: default buffer will be allocated
         -- buffer /= default_pointer implies buffer size = BUFSIZ
      do
         posix_setbuf (stream, buffer)
      end
   
   set_full_buffering (buffer: POINTER; size: INTEGER) is
         -- Determines buffering for a stream
         -- give NULL buffer so `setvbuf' will allocate a buffer
      local
         r: INTEGER
      do
         r := posix_setvbuf (stream, buffer, IOFBF, size)
         if r /= 0 then
            raise_posix_error
         end
      end
   
   set_line_buffering (buffer: POINTER; size: INTEGER) is
         -- Determines buffering for a stream
         -- give NULL buffer so `setvbuf' will allocate a buffer
      local
         r: INTEGER
      do
         r := posix_setvbuf (stream, buffer, IOLBF, size)
         if r /= 0 then
            raise_posix_error
         end
      end

   set_no_buffering is
         -- Turns off buffering
      local
         r: INTEGER
      do
         r := posix_setvbuf (stream, default_pointer, IONBF, 0)
         if r /= 0 then
            raise_posix_error
         end
      end
   

feature {NONE} -- read internal state
   
   eof_read: BOOLEAN
         -- set by `getc' and `gets', signals EOF read


feature -- read, C like

   last_byte: INTEGER
         -- last read character of `get_character'
         -- can be negative, so is more a last_shortint or so!
   
   getc, get_character is
         -- Reads a C unsigned char and converts it to an integer,
         -- the result is left in `last_byte'
         -- This function probably can be used to read a single
         -- byte
      do
         assert_input_ok
         last_byte := posix_fgetc (stream)
         eof_read := last_byte = const_EOF
      ensure
         eof_set: last_byte = const_EOF implies eof
      end

   gets, get_string (bytes: INTEGER) is
         -- Reads at most one less than `bytes' characters.
         -- No additional characters are read after a newline character
         -- or after end-of-file. If a newline character is read, it
         -- is returned too.
         -- Result is placed in `last_string'
      local
         p: POINTER
      do
         assert_input_ok
         assert_buf_has_room (bytes)
         p := posix_fgets (gets_buf.ptr, bytes, stream)
         if p = default_pointer then
            eof_read := posix_feof (stream)
            if not eof_read then
               raise_posix_error
            else
               last_string.clear
            end
         else
            sh.set_string_from_pointer (last_string, gets_buf.ptr)
            if eof then
               -- we don't want the eof to be True here, user has to call
               -- gets again to make Eof True, fits better with Eiffel loops
               clear_error
            end
            last_read := last_string.count
         end
      end
   
   
feature {NONE} -- gets private state
                         
   gets_buf: STDC_DYNAMIC_MEMORY
         -- buffer used by `gets'
                         
   assert_buf_has_room (needed_size: INTEGER) is
         -- make sure `gets_buf' has enough room
      do
         if gets_buf.size < needed_size then
            gets_buf.resize (needed_size)
         end
      end

   
feature -- read, Eiffel like   

   last_read: INTEGER
         -- last read bytes by some read_XXXX or get_string call
   
   last_character: CHARACTER
         -- last character read by getc

   last_string: STRING
         -- last string read by get_string

   read (buf: POINTER; bytes: INTEGER) is
      do
         do_read (buf, bytes)
      end

   read_character is
         -- read a single character and set `last_character'
         -- if end-of-file encountered, `eof' is True
      do
         get_character
         last_character := last_byte.to_character
         last_read := 1
      end

   read_string (bytes: INTEGER) is
         -- Read at most `n' characters, a value more expected by
         -- programmers not used to strings with a trailing byte.
         -- result is placed in `last_string'
         -- `last_string' includes the newline character!
      do
         get_string (bytes+1)
      end


feature -- write

   last_written: INTEGER
         -- last written bytes by some write_XXXX call
   
   put (any: ANY) is
      -- write class as string
      do
         puts (any.out)
      end
   
   putc (c: INTEGER) is
         -- write a single character
      local
         r: INTEGER
      do
         r := posix_fputc (c, stream)
         if r = const_EOF then
            raise_posix_error
         end
         need_flush := True
      ensure
         need_flush_set: need_flush
      end
   
   puts, put_string, write_string (s: STRING) is
         -- write a string
      require
         valid_string: s /= Void
      local
         cs: POINTER
         r: INTEGER
      do
         cs := sh.string_to_pointer(s)
         r := posix_fputs (cs, stream)
         if r = const_EOF then
            raise_posix_error
         end
         need_flush := True
      ensure
         need_flush_set: need_flush
      end

   ungetc (c: INTEGER) is
         -- pushes `c' back to the stream
         -- note that file positioning functions discard any
         -- pushed-back characters
      local
         r: INTEGER
      do
         -- and that's why we make sure position is not needed anymore
         assert_input_ok
         r := posix_ungetc (c, stream)
         if r = const_EOF then
            raise_posix_error
         end
      end

   write (buf: POINTER; bytes: INTEGER) is
         -- write `bytes' bytes from `buf'
      do
         do_write (buf, bytes)
      end
   

feature -- file position
   
   getpos, get_position: STDC_FILE_POSITION is
      -- get the current position, use `set_position' to return to 
      -- this saved position
      do
         create Result.make
         if posix_fgetpos (stream, Result.buf.ptr) /= 0 then
            raise_posix_error
         end
      end

   rewind is
      -- Sets the file position to the beginning of the file
      do
         posix_rewind (stream)
         eof_read := False
         need_flush := False
      ensure
         not_eof: not eof
         no_need_to_flush: not need_flush
      end
   
   seek (offset: INTEGER) is
         -- set file position to given absolute `offset'
      require
         valid_offset: offset >= 0
      do
         if posix_fseek (stream, offset, SEEK_SET) /= 0 then
            raise_posix_error
         end
         eof_read := False
         need_flush := False
      ensure
         not_eof: not eof
         no_need_to_flush: not need_flush
      end

   seek_from_current (offset: INTEGER) is
         -- set file position relative to current position
      do
         if posix_fseek (stream, offset, SEEK_CUR) /= 0 then
            raise_posix_error
         end
         eof_read := False
         need_flush := False
      ensure
         not_eof: not eof
         no_need_to_flush: not need_flush
      end
   
   seek_from_end (offset: INTEGER) is
         -- set file position relative to end of file
      require
         valid_offset: offset <= 0
      do
         if posix_fseek (stream, offset, SEEK_END) /= 0 then
            raise_posix_error
         end
         eof_read := False
         need_flush := False
      ensure
         not_eof: not eof
         no_need_to_flush: not need_flush
      end

   setpos, set_position (a_position: STDC_FILE_POSITION) is
         -- set the current position
      require
         valid_position: a_position /= Void
      do
         if posix_fsetpos (stream, a_position.buf.ptr) /= 0 then
            raise_posix_error
         end
         eof_read := False
         need_flush := False
      ensure
         not_eof: not eof
         no_need_to_flush: not need_flush
      end

   tell: INTEGER is
         -- The current position
      do
         Result := posix_ftell (stream)
      end
   

feature -- other
   
   clearerr, clear_error is
         -- Clears end-of-file and error indicators for a stream
      do
         posix_clearerr (stream)
      end
   
   
feature -- queries
   
   eof: BOOLEAN is
         -- True if eof encountered by getc or,
         -- if the end-of-file indicator is set
      do
         Result := eof_read or else posix_feof (stream)
      end

   error: BOOLEAN is
         -- True if and only if the error indicator is set
      do
         Result := posix_ferror (stream)
      end

   filename: STRING is
         -- the filename of this file
      do
         Result := portable_path
      end

   is_open: BOOLEAN is
      do
         Result := stream /= default_pointer
      end
   
   mode: STRING
         -- mode in which the file is opened/created
   
   
feature {NONE} -- internal
   
   dispose is
      do
         if is_open then
            close
         end
      end

   do_make is
         -- make sure invariants are satisfied
      do
         make_path
         if last_string = Void then
            create last_string.make (256)
         end
         if gets_buf = Void then
            create gets_buf.allocate_and_clear (256)
         end
      end
   
   do_open (a_path, a_mode: STRING) is
      require
         -- `a_mode' is a valid mode
      local
         cpath: POINTER
         cmode: POINTER
      do
         portable_path.make_expand (a_path)
         cpath := sh.string_to_pointer (portable_path)
         set_mode (a_mode)
         cmode := sh.string_to_pointer (mode)
         stream := posix_fopen (cpath, cmode)
         if not is_open then
            raise_posix_error
         end
      ensure
         file_is_open: is_open
      end

   do_read (buf: POINTER; bytes: INTEGER) is
      require
         is_opened: is_open
      do
         assert_input_ok
         last_read := posix_fread (buf, 1, bytes, stream)
         if last_read /= bytes then
            raise_posix_error
         end
      ensure
         successfull_read: last_read = bytes
      end

   do_write (buf: POINTER; bytes: INTEGER) is
      require
         is_opened: is_open
         valid_bytes: bytes >= 0
      do
         last_written := posix_fwrite (buf, 1, bytes, stream)
         if last_written /= bytes then
            raise_posix_error
         end
         need_flush := True
      ensure
         successfull_write: last_written = bytes
         need_flush_set: need_flush
      end
   
   need_flush: BOOLEAN
         -- need to be set after any write operation
         -- reset after any flush operation
   
   assert_input_ok is      
         -- ANSI C requires that after a read a file positioning function
         -- or flush is performed
      do
         if need_flush then
            seek_from_current (0)
         end
      ensure
         no_need_to_flush: not need_flush
      end
   
   set_mode (a_mode: STRING) is
         -- use this routine to set the mode variable. It takes care 
         -- of adding the binary or text flag in its descendents
      require
         valid_mode: a_mode /= Void and then not a_mode.is_empty
      deferred
      end


feature {BASE_FILE_DESCRIPTOR} -- state  
   
   stream: POINTER
   
   
invariant
   
   last_string_valid: last_string /= Void
   gets_buf_valid: gets_buf /= Void

end -- class STDC_FILE
