indexing

   description: "Class that generates an XML string."

   notes: "This is not DOM, but a kind of streaming XML generation."

   author: "Berend de Boer"
   date: "$Date: Fri, 13 Oct 2000 13:36:48 +0200 $"
   revision: "$Revision: 1.3 $"

class 

   XML_GENERATOR


creation
   
   make
   
   
feature -- creation

   make is
      do
         create my_xml.make (512)
         create tags.make (0, 32)
         create attributes.make (0, 32)
         create values.make (0, 32)
         clear
      end
   

feature -- constants from the XML specification, should be Unicode...
   
   ValidFirstChars: STRING is "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_:"
         -- which characters are valid as the first character
   
   ValidOtherChars: STRING is "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_:.-0123456789"
         -- which characters are valid as second etc characters
   

feature -- queries

   is_header_written: BOOLEAN is
      do
         Result := header_written
      end
   
   is_started (tag: STRING): BOOLEAN is
      do
         Result := top_tag.is_equal (tag)
      end
   
   is_tag_started: BOOLEAN is
      do
         Result := not is_empty
      end

   is_valid_attribute_name (attribute: STRING): BOOLEAN is
         -- Return True if this is a valid attribute name
      local
         first,
         other: CHARACTER
         i: INTEGER
      do
         Result := attribute /= Void and then not attribute.is_empty
         if Result then
            first := attribute.item (1)
            Result := ValidFirstChars.has (first)
            from
               i := 2
            until
               not Result or else i = attribute.count
            loop
               other := attribute.item (i)
               Result := ValidOtherChars.has (other)
               i := i + 1
            end
         end
      end
   
   xml: STRING is
         -- the result
      do
         Result := my_xml
      end
   

feature -- influence state
   
   clear is
         -- start fresh
      do
         xml.clear
         tags.clear_all
         top := 0
         clear_attributes
         header_written := False
         new_line_written := True
         may_close_tag := False
         tag_closed := False
         tag_written := True
      ensure
         no_tags: is_empty
      end


feature -- commands that expand `xml'

   add_header is
      require
         valid_point_for_header: not is_header_written
      do
         extend ("<?xml version=%"1.0%" encoding=%"ISO-8859-1%" ?>")
         new_line
         header_written := True
      end

   add_data (data: STRING) is      
         -- write data in the current tag
      require
         valid_point_for_data: is_tag_started
      do
         assure_last_tag_written
         extend (data)
      end

    add_tag (tag, data: STRING) is
         -- shortcut for `add_tag', `add_data' and `stop_tag'
      require
         have_header: is_header_written
      do
         start_tag (tag)
         add_data (data)
         stop_tag
      end

   extend (stuff: STRING) is
         -- add anything to the current `xml' string, you're on your own here!
      do
         if stuff /= Void then
            my_xml.append_string (stuff)
         end
      end

   new_line is
      do
         extend ("%R%N")
         new_line_written := True
      end
   
   set_attribute (attribute, value: STRING) is
         -- set an attribute of the current tag
      require
         valid_attribute: is_valid_attribute_name (attribute)
      do
         do_set_attribute (attribute, make_valid_attribute_value (value))
      end
   
   start_tag (tag: STRING) is
         -- start a new tag
      do
         assure_last_tag_written
         push_tag (tag)
         tag_written := False
      end

   stop_tag is
         -- stop last started tag
      require
         tag_is_started: is_tag_started
      do
         may_close_tag := True
         assure_last_tag_written
         if not tag_closed then
            if new_line_written then
               indent
            end
            extend ("</")
            extend (top_tag)
            extend (">")
         end
         pop_tag
         new_line
      end


feature {NONE} -- internal xml change

   assure_last_tag_written is
      local
         i: INTEGER
         v: STRING
      do
         tag_closed := False
         if not tag_written then
            if not new_line_written then
               new_line
            end
            indent
            extend ("<")
            extend (top_tag)
            from
               i := 0
            until
               i = attribute_count
            loop
               extend (" ")
               extend (attributes.item (i))
               extend ("=%"")
               v := values.item (i)
               if v /= Void then
                  extend (v)
               end
               extend ("%"")
               i := i + 1
            end
            clear_attributes
            if may_close_tag then
               extend ("/>")
               tag_closed := True
            else
               extend (">")
            end
            tag_written := True
         end
         may_close_tag := False
      end
   
   indent is
      local
         spaces: STRING
      do
         -- create spaces.make_filled (' ', 2 * (tag_count - 1 ) )
         if tag_count > 1 then
            create spaces.make ( 2 * (tag_count - 1 ) )
            spaces.extend_multiple (' ', 2 * (tag_count - 1 ) )
            extend (spaces)
         end
         new_line_written := False
      end

   make_valid_attribute_value (value: STRING): STRING is
      local
         i: INTEGER
      do
         Result := clone (value)
         if Result /= Void then
            from
               i := 1
            until
               i > Result.count
            loop
               inspect Result.item (i)
               when '"' then
                  Result.remove (i)
                  insert_string (Result, i, QuoteQuote)
                  i := i + QuoteQuote.count
               when '<' then
                  Result.remove (i)
                  insert_string (Result, i, QuoteLt)
                  i := i + QuoteLt.count
               when '>' then
                  Result.remove (i)
                  insert_string (Result, i, QuoteGt)
                  i := i + QuoteGt.count
               when '&' then
                  Result.remove (i)
                  insert_string (Result, i, QuoteAmp)
                  i := i + QuoteAmp.count
               else
                  i := i + 1
               end
            end
         end
      end
   

feature {NONE} -- stack of tags, oh we miss a portable EiffelBase so
   
   tags: ARRAY[STRING]

   top: INTEGER

   is_empty: BOOLEAN is
      do
         Result := top = 0
      end
   
   pop_tag is
      require
         has_tags: not is_empty
      do
         top := top - 1
      end

   push_tag (tag: STRING) is
      do
         tags.put (tag, top)
         top := top + 1
      ensure
         have_tags: not is_empty
      end
   
   tag_count: INTEGER is
         -- number of entries on stack
      do
         Result := top
      end
   
   top_tag: STRING is
      require
         has_tags: not is_empty
      do
         Result := tags.item (top - 1)
      end


feature {NONE} -- tag attributes   
   
   attributes: ARRAY[STRING]
   
   values: ARRAY[STRING]

   current_attribute: INTEGER

   attribute_count: INTEGER
   
   add_attribute (attribute: STRING) is
      do
         if attributes.count = attribute_count then
            attributes.resize (attributes.lower, attributes.upper * 2)
            values.resize (values.lower, values.upper * 2)
         end
         attributes.put (attribute, attribute_count)
         values.put (Void, attribute_count)
         attribute_count := attribute_count + 1
      end

   clear_attributes is
      do
         attributes.clear_all
         values.clear_all
         attribute_count := 0
      end
   
   do_set_attribute (attribute, value: STRING) is
      do
         if exist_attribute (attribute) then
            put_value (value)
         else
            add_attribute (attribute)
            put_value (value)
         end
      end

   exist_attribute (attribute: STRING): BOOLEAN is
      local
         Found: BOOLEAN
      do
         from
            current_attribute := 0
            Found := False
         until
            Found or else current_attribute = attribute_count
         loop
            Found := attributes.item (current_attribute).is_equal (attribute)
            if not Found then
               current_attribute := current_attribute + 1
            end
         end
      ensure
         Result implies valid_current_attribute
      end

   put_value (value: STRING) is
      require
         valid_index: valid_current_attribute
      do
         values.put (value, current_attribute)
      end

   valid_current_attribute: BOOLEAN is
      do
         Result := 
            current_attribute >= 0 and 
            current_attribute < attribute_count
      end


feature {NONE} -- private state
   
   my_xml: STRING
         -- the string builded as we go

   header_written: BOOLEAN

   may_close_tag: BOOLEAN
   
   new_line_written: BOOLEAN
   
   tag_closed: BOOLEAN

   tag_written: BOOLEAN
   
   
feature {NONE} -- attribute value sanitize
   
   QuoteAmp: STRING is "&amp;"
   QuoteLt: STRING is "&lt;"
   QuoteGt: STRING is "&gt;"
   QuoteQuote: STRING is "&quot;"
   
   insert_string (res: STRING; index: INTEGER; substr: STRING) is
      local
         i: INTEGER
      do
         from
            i := 1
         until
            i > substr.count
         loop
            res.insert (substr.item (i), index + i - 2)
            i := i + 1
         end
      end

invariant

   same_size: attributes.count = values.count

end -- class XML_GENERATOR
